import numpy as np

import matplotlib.pyplot as plt


def dh_angles_hist(dh_angles):
    """
    Histograms of dihedral angles in tetrahedrons
    """

    n, bins, patches = plt.hist(
        x=dh_angles, bins="auto", color="#0504aa", alpha=0.7, rwidth=0.85
    )
    maxN = np.amax(n)
    minDh = np.amin(dh_angles)
    maxDh = np.amax(dh_angles)
    plt.plot(np.linspace(minDh, minDh, 10), np.linspace(0, maxN / 2, 10), "r-")
    plt.plot(np.linspace(maxDh, maxDh, 10), np.linspace(0, maxN / 2, 10), "r-")
    plt.text(minDh, maxN, minDh)
    plt.text(maxDh, maxN, maxDh)
    frame1 = plt.gca()
    frame1.axes.yaxis.set_ticklabels([])
    frame1.axes.yaxis.set_ticks([])
    # Hide the right and top spines
    frame1.spines["right"].set_visible(False)
    frame1.spines["top"].set_visible(False)
    plt.show()


def re_ratios_hist(re_ratios):
    """
    Histograms of radius-edge metric
    """
    n, bins, patches = plt.hist(
        x=re_ratios, bins="auto", color="#0504aa", alpha=0.7, rwidth=0.85
    )
    maxN = np.amax(n)
    minRe = np.amin(re_ratios)
    maxRe = np.amax(re_ratios)
    plt.plot(np.linspace(minRe, minRe, 10), np.linspace(0, maxN / 2, 10), "r-")
    plt.plot(np.linspace(maxRe, maxRe, 10), np.linspace(0, maxN / 2, 10), "r-")
    plt.text(minRe, maxN, minRe)
    plt.text(maxRe, maxN, maxRe)
    frame1 = plt.gca()
    frame1.axes.yaxis.set_ticklabels([])
    frame1.axes.yaxis.set_ticks([])
    # Hide the right and top spines
    frame1.spines["right"].set_visible(False)
    frame1.spines["top"].set_visible(False)
    plt.show()



import numpy as np
import meshio
from mpi4py import MPI

import SeismicMesh

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

hmin = 0.10
bbox = (-1.0, 1.0, -1.0, 1.0, -1.0, 1.0)


# signed distancem function
def fd(p):
    return np.sqrt((p ** 2).sum(1)) - 1.0


# sizing function
def fh(p):
    # note for parallel execution this logic is required
    # since the decomposition of the sizing function passes a tuple to fh
    if type(p) == tuple:
        h = np.zeros_like(p[0]) + hmin
    else:
        h = np.zeros_like(p) + hmin
    return h


# Construct mesh generator
mshgen = SeismicMesh.MeshGenerator(hmin=hmin, bbox=bbox, fd=fd, fh=fh)

# Build the mesh
points, cells = mshgen.build(max_iter=75, axis=1)

# Remove degenerate slivers
points, cells = mshgen.build(points=points, mesh_improvement=True)

if rank == 0:
    meshio.write_points_cells(
        "cylinder_3d.vtk",
        points / 1000,
        [("tetra", cells)],
        file_format="vtk",
    )



import numpy as np
import zipfile
import meshio

import SeismicMesh
from SeismicMesh.geometry import (
    SignedDistanceFunctionGenerator as SdfGen,
)  # import the tool used to generate the SDF

# This example highlights how build a 3D mesh from a signed distance function built from an iso-contour in serial.

# The velocity model was downloaded from here:
# https://s3.amazonaws.com/open.source.geoscience/open_data/seg_eage_models_cd/Salt_Model_3D.tar.gz

# Dimensions of model
nx, ny, nz = 676, 676, 210

path = "velocity_models/Salt_Model_3D/3-D_Salt_Model/VEL_GRIDS/"
# Extract Saltf@@ from SALTF.ZIP
zipfile.ZipFile(path + "SALTF.ZIP", "r").extract("Saltf@@", path=path)

# Load data
with open(path + "Saltf@@", "r") as file:
    vp = np.fromfile(file, dtype=np.dtype("float32").newbyteorder(">"))
    vp = vp.reshape(nx, ny, nz, order="F")
    vp = np.flipud(vp.transpose((2, 0, 1)))  # z, x and then y


# Bounding box describing domain extents (corner coordinates)
bbox = (-4200, 0, 0, 13520, 0, 13520)

# Minimum mesh resolution
hmin = 100.0

ef = SeismicMesh.MeshSizeFunction(
    bbox=bbox,
    velocity_grid=vp,
    dt=0.001,
    freq=2,
    wl=5,
    grade=0.25,
    hmin=hmin,
    hmax=5e3,
)

# Build the sizing function
ef.build()

# Bulid a signed distance function from the geodataset
# Only capturing features with a minimum velocity greater than the speed of sound in water
# and less than the maximum velocity in the model
SDF = SdfGen(
    bbox=bbox,
    field=vp,
    min_threshold=3500,
    gridspacing=(20.0, 20.0, 20.0),
    narrow=1000,
).SDF

# Construct mesh generator
mshgen = SeismicMesh.MeshGenerator(hmin=hmin, bbox=bbox, fd=SDF, fh=ef.fh)

# Build the mesh
points, cells = mshgen.build(max_iter=50)

# Mesh improvement
points, cells = mshgen.build(points=points, mesh_improvement=True)

# Write the mesh to disk
meshio.write_points_cells(
    "eage_salt_body.vtk", points / 1000.0, [("tetra", cells)], file_format="vtk"
)



from .migration import (
    enqueue,
    aggregate,
    exchange,
    localize_points,
    localize_sizing_function,
)


__all__ = [
    "enqueue",
    "aggregate",
    "exchange",
    "localize_points",
    "localize_sizing_function",
]



import numpy as np


def blocker(points, rank, nblocks, axis=0):
    """Decompose point coordinates into # of blocks
    Blocks are orientated parallel to axis and have two neighbors
    above and below/left and right to the block depending on axis.
    """
    num_points, dim = points.shape
    EPS = np.finfo(float).eps

    assert dim > 2 or dim < 3, "dimensions of points are wrong"
    assert num_points // nblocks > 1, "too few points for chosen nblocks"

    xlims = points[:, 0].min() - EPS, points[:, 0].max() + EPS
    ylims = points[:, 1].min() - EPS, points[:, 1].max() + EPS
    if dim > 2:
        zlims = points[:, 2].min() - EPS, points[:, 2].max() + EPS

    if axis == 0:
        if dim == 2:
            xx, yy = np.meshgrid(
                np.linspace(*xlims, 1, endpoint=False),
                np.linspace(*ylims, nblocks, endpoint=False),
            )

            dx = (xlims[1] - xlims[0]) / 1
            dy = (ylims[1] - ylims[0]) / nblocks
        elif dim == 3:
            xx, yy, zz = np.meshgrid(
                np.linspace(*xlims, 1, endpoint=False),
                np.linspace(*ylims, nblocks, endpoint=False),
                np.linspace(*zlims, 1, endpoint=False),
            )

            dx = (xlims[1] - xlims[0]) / 1
            dy = (ylims[1] - ylims[0]) / nblocks
            dz = (zlims[1] - zlims[0]) / 1

    elif axis == 1:
        if dim == 2:
            xx, yy = np.meshgrid(
                np.linspace(*xlims, nblocks, endpoint=False),
                np.linspace(*ylims, 1, endpoint=False),
            )

            dx = (xlims[1] - xlims[0]) / nblocks
            dy = (ylims[1] - ylims[0]) / 1
        elif dim == 3:
            xx, yy, zz = np.meshgrid(
                np.linspace(*xlims, nblocks, endpoint=False),
                np.linspace(*ylims, 1, endpoint=False),
                np.linspace(*zlims, 1, endpoint=False),
            )

            dx = (xlims[1] - xlims[0]) / nblocks
            dy = (ylims[1] - ylims[0]) / 1
            dz = (zlims[1] - zlims[0]) / 1
    elif axis == 2:
        xx, yy, zz = np.meshgrid(
            np.linspace(*xlims, 1, endpoint=False),
            np.linspace(*ylims, nblocks, endpoint=False),
            np.linspace(*zlims, nblocks, endpoint=False),
        )

        dx = (xlims[1] - xlims[0]) / 1
        dy = (ylims[1] - ylims[0]) / 1
        dz = (zlims[1] - zlims[0]) / nblocks

    blocks = []
    block_extents = []

    if dim == 2:
        for low_x, low_y in zip(xx.ravel(), yy.ravel()):
            block = points[
                (points[:, 0] >= low_x)
                & (points[:, 0] <= low_x + dx)
                & (points[:, 1] >= low_y)
                & (points[:, 1] <= low_y + dy)
            ]
            if block.shape[0]:
                blocks.append(block)

        block_extents = []
        for block in blocks:
            tmpm = np.amin(block, axis=0)
            tmpp = np.amax(block, axis=0)
            # min x min y max x max y
            block_extents.append([tmpm[0], tmpm[1], tmpp[0], tmpp[1]])
    elif dim == 3:
        for low_x, low_y, low_z in zip(xx.ravel(), yy.ravel(), zz.ravel()):
            block = points[
                (points[:, 0] >= low_x)
                & (points[:, 0] <= low_x + dx)
                & (points[:, 1] >= low_y)
                & (points[:, 1] <= low_y + dy)
                & (points[:, 2] >= low_z)
                & (points[:, 2] <= low_z + dz)
            ]
            if block.shape[0]:
                blocks.append(block)

        block_extents = []
        for block in blocks:
            tmpm = np.amin(block, axis=0)
            tmpp = np.amax(block, axis=0)
            # min x min y min z max x max y max z
            block_extents.append([tmpm[0], tmpm[1], tmpm[2], tmpp[0], tmpp[1], tmpp[2]])

    return blocks, block_extents



import SeismicMesh


def example_write():
    # Name of SEG-Y file containg velocity model.
    fname = "velocity_models/vel_z6.25m_x12.5m_exact.segy"
    vp = SeismicMesh.ReadSegy(fname)
    bbox = (-12e3, 0, 0, 67e3)

    # Construct mesh sizing object from velocity model
    ef = SeismicMesh.MeshSizeFunction(
        bbox=bbox,
        velocity_grid=vp,
        domain_ext=2e3,
        dt=0.001,
        freq=5,
        wl=5,
        hmax=1e3,
        hmin=50.0,
        grade=0.05,
    )

    ef = ef.build()

    # Write it to an hdf5 file for later use.
    ef.WriteVelocityModel("BP2004")

    # Save your mesh size function options.
    ef.SaveMeshSizeFunctionOptions("BP2004_SizingFunction")


if __name__ == "__main__":

    example_write()



from mpi4py import MPI
import meshio

import SeismicMesh

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()


def example_2D():
    """
    Build a mesh of the BP2004 benchmark velocity model in serial or parallel
    Takes roughly 1 minute with 2 processors.
    The velocity model can be downloaded from here: https://wiki.seg.org/wiki/2004_BP_velocity_estimation_benchmark_model
    """

    # Name of SEG-Y file containg velocity model.
    fname = "velocity_models/vel_z6.25m_x12.5m_exact.segy"
    # Read in it
    vp = SeismicMesh.ReadSegy(fname)

    # Bounding box describing domain extents (corner coordinates)
    bbox = (-12e3, 0, 0, 67e3)

    # Construct mesh sizing object from velocity model
    ef = SeismicMesh.MeshSizeFunction(
        bbox=bbox,
        velocity_grid=vp,
        freq=2,
        wl=10,
        dt=0.001,
        hmin=75.0,
        grade=0.15,
        domain_ext=1e3,
        padstyle="linear_ramp",
    )

    # Build mesh size function
    ef = ef.build()

    # Write to disk for later use
    ef.WriteVelocityModel("BP2004_w1KM_EXT")

    # Visualize the mesh size function
    ef.plot()

    # Construct a mesh generator object
    mshgen = SeismicMesh.MeshGenerator(ef)

    # Build the mesh
    points, facets = mshgen.build(axis=1)

    if rank == 0:

        # Write the mesh as a vtk format for visualization in Paraview
        meshio.write_points_cells(
            "BP2004.vtk",
            points / 1000,
            [("triangle", facets)],
            file_format="vtk",
        )
        # Write to gmsh22 format (quite slow) used by many numerical solvers
        meshio.write_points_cells(
            "BP2004.msh",
            points / 1000,
            [("triangle", facets)],
            file_format="gmsh22",
            binary=False,
        )


if __name__ == "__main__":

    example_2D()



from .mesh_generator import MeshGenerator

from .cpp.delaunay_class import DelaunayTriangulation
from .cpp.delaunay_class3 import DelaunayTriangulation3

from .cpp.c_cgal import (
    delaunay2,
    delaunay3,
    circumballs2,
    circumballs3,
)


__all__ = [
    "DelaunayTriangulation",
    "DelaunayTriangulation3",
    "delaunay2",
    "delaunay3",
    "circumballs2",
    "circumballs3",
    "MeshGenerator",
]



import numpy as np
import meshio
from mpi4py import MPI

import SeismicMesh

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

hmin = 0.10
bbox = (-1.0, 1.0, -1.0, 1.0, -1.0, 1.0)

# This examples builds a small and simple cylinder in either serial or parallel.


def fd(p):
    """sizing function of a cylinder in [-1., -1., -.1] x [1., 1., 1.]"""
    r, z = np.sqrt(p[:, 0] ** 2 + p[:, 1] ** 2), p[:, 2]
    d1, d2, d3 = r - 1.0, z - 1.0, -z - 1.0
    d4, d5 = np.sqrt(d1 ** 2 + d2 ** 2), np.sqrt(d1 ** 2 + d3 ** 2)
    d = np.maximum.reduce([d1, d2, d3])
    ix = (d1 > 0) * (d2 > 0)
    d[ix] = d4[ix]
    ix = (d1 > 0) * (d3 > 0)
    d[ix] = d5[ix]
    return d


# sizing function
def fh(p):
    # note for parallel execution this logic is required
    # since the decomposition of the sizing function passes a tuple to fh
    if type(p) == tuple:
        h = np.zeros_like(p[0]) + hmin
    else:
        h = np.zeros_like(p) + hmin
    return h


# Construct mesh generator
mshgen = SeismicMesh.MeshGenerator(hmin=hmin, bbox=bbox, fd=fd, fh=fh)

# Build the mesh
points, cells = mshgen.build(max_iter=75)

# Remove degenerate slivers
points, cells = mshgen.build(points=points, mesh_improvement=True)

if rank == 0:
    meshio.write_points_cells(
        "cylinder_3d.vtk",
        points / 1000,
        [("tetra", cells)],
        file_format="vtk",
    )



import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as a3


def plot_points3(points, marker="gx", hold_on=False):
    """
    plot points in 3d
    """
    axes = a3.Axes3D(plt.figure())
    axes.plot(points[:, 0], points[:, 1], points[:, 2], marker)
    axes.set_axis_off()
    if hold_on is False:
        plt.show()
    return None



import numpy as np
from mpi4py import MPI

import meshio

import SeismicMesh

from SeismicMesh.geometry import (
    SignedDistanceFunctionGenerator as SdfGen,
)  # import the tool used to generate the SDF


comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

if size > 1:
    print("This is a serial example!")


def example_2D_Irregular():
    """
    Build a 2D mesh of a benchmark velocity  model of a Monuntainous thrust region.
    Builds in serial or parallel. Demonstrates how to ensure the topography is respected
    in the mesh.

    Velocity model was downloaded from here: https://wiki.seg.org/wiki/1994_BP_migration_from_topography
    """

    # Name of SEG-Y file containg velocity model.
    fname = "velocity_models/velocity.segy"
    vp = SeismicMesh.ReadSegy(fname)

    # Bounding box describing domain extents (corner coordinates)
    bbox = (-8e3, 2e3, 0, 25e3)
    hmin = 37.5

    # Construct mesh sizing object from velocity model
    ef = SeismicMesh.MeshSizeFunction(
        bbox=bbox,
        velocity_grid=vp,
        freq=4,
        wl=10,
        dt=0.001,
        hmin=hmin,
        grade=0.15,
    )

    # Build mesh size function
    ef = ef.build()

    # Write to disk for later use
    ef.WriteVelocityModel("Irregular2D")

    # Visualize the mesh size function
    ef.plot()

    # Bulid a signed distance function from the seismic velocity model
    # Some pockets of velocity < 4000 exist, fill those in.
    vp2 = vp.copy()
    vp2 = np.where(vp2 < 4000, 4001, vp2)
    SDF = SdfGen(
        bbox=bbox,
        field=vp2,
        min_threshold=4000.0,
        gridspacing=(10.0, 15.0),
    ).SDF

    # Construct a mesh generator object
    mshgen = SeismicMesh.MeshGenerator(bbox=bbox, hmin=hmin, fh=ef.fh, fd=SDF)

    # Build the mesh
    points, facets = mshgen.build(perform_checks=True)

    if rank == 0:

        # Write the mesh as a vtk format for visualization in Paraview
        meshio.write_points_cells(
            "Irregular2D.vtk",
            points / 1000,
            [("triangle", facets)],
            file_format="vtk",
        )
        # Write to gmsh22 format (quite slow) used by many numerical solvers
        meshio.write_points_cells(
            "Irregular2D.msh",
            points / 1000,
            [("triangle", facets)],
            file_format="gmsh22",
            binary=False,
        )


if __name__ == "__main__":

    example_2D_Irregular()



from .signed_distance_functions import dblock, drectangle, dintersect
from .utils import (
    calc_re_ratios,
    doAnyOverlap,
    linter,
    laplacian2,
    vertex_to_entities,
    remove_external_entities,
    unique_rows,
    fixmesh,
    simpvol,
    simpqual,
    get_centroids,
    get_edges,
    get_facets,
    get_boundary_entities,
    get_boundary_vertices,
    get_boundary_edges,
    get_winded_boundary_edges,
    get_boundary_facets,
    delete_boundary_entities,
    vtInEntity3,
)
from .utils import SignedDistanceFunctionGenerator
from .cpp.fast_geometry import (
    calc_volume_grad,
    calc_circumsphere_grad,
    calc_3x3determinant,
    calc_4x4determinant,
    calc_dihedral_angles,
)

__all__ = [
    "SignedDistanceFunctionGenerator",
    "calc_re_ratios",
    "calc_volume_grad",
    "calc_circumsphere_grad",
    "calc_3x3determinant",
    "calc_4x4determinant",
    "dblock",
    "drectangle",
    "dintersect",
    "calc_dihedral_angles",
    "doAnyOverlap",
    "linter",
    "laplacian2",
    "vertex_to_entities",
    "remove_external_entities",
    "unique_rows",
    "fixmesh",
    "simpvol",
    "simpqual",
    "get_centroids",
    "get_edges",
    "get_facets",
    "get_boundary_vertices",
    "get_boundary_entities",
    "delete_boundary_entities",
    "get_boundary_edges",
    "get_boundary_facets",
    "get_winded_boundary_edges",
    "vtInEntity3",
]
